import subprocess
from report.report_utils import *
from report.report_recommendation import *


def generate_report(customer_name,stam_name,review_date,data_folder,output_folder):
    if os.path.exists(output_folder):
        print("WARN: output folder exist")
        exit(0)
    else:
        os.mkdir(output_folder)

    report_file=output_folder + "/opensearch_report.md"
    print("generating report in: " + report_file)

    if os.path.exists(data_folder):
        print("generating report from the data in the folder: " + data_folder)
    else:
        print("data folder does not exist: " + data_folder)
        exit(1)


    output_header(customer_name, stam_name, review_date,report_file,output_folder)

    recommendations = generate_recommendation(data_folder, report_file)
    output_recommendations_as_csv(recommendations, output_folder)
    output_summary(read_recommendations_from_csv(output_folder),report_file)
    output_body(data_folder, recommendations, report_file)

    # General recommendation with static content
    output_static_recommendation(data_folder, report_file)
    # Appendix
    output_appendix(data_folder, report_file)
    # Reference
    output_reference(report_file)



def output_header(customer_name,stam_name,review_date,report_file,output_folder):
    copy_cmd = "cp report/AWS-Operation-Review.png " + output_folder + "/AWS-Operation-Review.png"
    subprocess.call(copy_cmd, shell=True)
    write_line(report_file,"![AWS Operation Review](AWS-Operation-Review.png)")
    write_line(report_file,"# Amazon OpenSearch Review for " + customer_name)
    write_line(report_file,"* Enterprise Support Specalist: " + stam_name)
    write_line(report_file,"* Review Date: " +review_date)
    write_line(report_file, "")
    write_line(report_file, "")


def output_summary(recs,report_file):
    write_line(report_file, "## Summary")
    write_line(report_file, "")
    write_line(report_file, "Key Recommendations identified on your cluster:")
    write_line(report_file, "")
    write_line(report_file, "|  No  |                        Observation                        |    Impact    |  Link   |")
    write_line(report_file, "|:----:| :-------------------------------------------------------: | :----------: | :------:|")
    row_number = 1
    for index, row in recs.iterrows():
        line = "| {} | {} | {} | [{}](##{}) |".format(row_number,row["observation"], row["impact"], row["report_section"],row["report_section"])
        write_line(report_file,line)
        row_number +=1
    write_line(report_file, "")


def output_body(data_folder,recommendations, report_file):
    output_cluster(data_folder,recommendations,report_file)
    output_node(data_folder,recommendations, report_file)
    output_index(data_folder,recommendations, report_file)
    output_shard(data_folder,recommendations, report_file)


def output_static_recommendation(data_folder,report_file):
    write_line(report_file, "")
    config_file = os.path.join(report.__path__[0], "general_recommendation.md")
    with open(config_file) as f:
        rec = f.read()
        write_line(report_file,rec)

def output_appendix(data_folder,report_file):
    write_line(report_file, "## Appendix")
    write_line(report_file,"")
    output_index_table(data_folder, report_file)
    output_shards_table(data_folder, report_file)
    output_allocation_table(data_folder, report_file)
    output_thread_pool_table(data_folder, report_file)

def output_reference(report_file):
    write_line(report_file, "## Reference")
    write_line(report_file, "(informational)")
    write_line(report_file, "")
    write_line(report_file, "[Best practices for Amazon OpenSearch Service](https://docs.aws.amazon.com/opensearch-service/latest/developerguide/bp.html)")
    write_line(report_file, "")
    write_line(report_file, "[Monitoring Amazon OpenSearch Service domains](https://docs.aws.amazon.com/opensearch-service/latest/developerguide/monitoring.html)")
    write_line(report_file, "")
    write_line(report_file, "[Increase Amazon Elasticsearch Service performance by upgrading to Graviton2](https://aws.amazon.com/blogs/big-data/increase-amazon-elasticsearch-service-performance-by-upgrading-to-graviton2/)")
    write_line(report_file, "")



def output_cluster(data_folder, recommendations, report_file):
    write_line(report_file, "## Cluster")
    write_line(report_file, "")
    output_cluster_domain_config(data_folder, recommendations,report_file)
    output_cluster_settings_and_state(data_folder, recommendations,report_file)
    output_cluster_health_and_stats(data_folder,recommendations,report_file)

    write_line(report_file, "")
    write_line(report_file, "### Recommendations")
    write_line(report_file, "")
    output_body_section(recommendations, report_file, "cluster")
    write_line(report_file, "")


def output_cluster_domain_config(data_folder, recommendations,report_file):
    domain_recommendations, domain_config = gen_domain_config(data_folder, report_file)
    write_line(report_file, "### Cluster Infrastructure Information:")
    write_line(report_file, "")
    write_line(report_file,
               "| Category |  Setting  |                        Value                        |")
    write_line(report_file,
               "|:------: | :----:| :-------------------------------------------------------: |")
    write_line(report_file, "| General | Version | " +domain_config.version + " |")
    if domain_config.dedicated_master_enabled:
        write_line(report_file, "| General | Dedicated Master Enabled | " + str(domain_config.dedicated_master_enabled) + " |")
    else:
        write_line(report_file,
                   "| General | Dedicated Master Enabled | <span style=\"color:red\">**" + str(domain_config.dedicated_master_enabled) + "**</span> |")
    write_line(report_file, "| General | Master Type | " + domain_config.dedicated_master_type + " |")
    write_line(report_file, "| General | Master Count | " + str(domain_config.dedicated_master_count) + " |")
    if domain_config.zone_awareness_enabled:
        write_line(report_file, "| General | Zone Awareness Enabled | " + str(domain_config.zone_awareness_enabled) + " |")
    else:
        write_line(report_file,
                   "| General | Zone Awareness Enabled | <span style=\"color:red\">**" + str(domain_config.zone_awareness_enabled) + "**</span> |")
    if domain_config.dn_zone_number < 3:
        write_line(report_file, "| General | Zone Number | <span style=\"color:red\">**" + str(domain_config.dn_zone_number) + "**</span> |")
    else:
        write_line(report_file, "| General | Zone Number | " + str(domain_config.dn_zone_number) + " |")
    write_line(report_file, "| General | Data Nodes Type | " + domain_config.instance_type + " |")
    write_line(report_file, "| General | Data Nodes Count | " + str(domain_config.instance_count) + " |")
    write_line(report_file, "| General | EBS Volume Type | " + domain_config.ebs_volume_type + " |")
    write_line(report_file, "| General | EBS Volume Size | " + str(domain_config.ebs_volume_size) + " |")
    if domain_config.warm_enabled:
        write_line(report_file, "| Cost | UltraWarm Enabled | " + str(domain_config.warm_enabled) + " |")
        write_line(report_file, "| Cost | UltraWarm Instance Type | " + str(domain_config.warm_instance_type) + " |")
        write_line(report_file, "| Cost | UltraWarm Instance Count | " + str(domain_config.warm_instance_count) + " |")
    else:
        write_line(report_file, "| Cost | UltraWarm Enabled | <span style=\"color:red\">**" + str(domain_config.warm_enabled) + "**</span> |")
    if domain_config.code_enabled:
        write_line(report_file, "| Cost | Code Storage Enabled | " + str(domain_config.code_enabled) + " |")
    else:
        write_line(report_file, "| Cost | Code Storage Enabled | <span style=\"color:red\">**" + str(domain_config.code_enabled) + "**</span> |")

    if domain_config.auto_tune_enable:
        write_line(report_file, "| Performance | Auto Tune Enabled | " + str(domain_config.auto_tune_enable) + " |")
    else:
        write_line(report_file, "| Performance | Auto Tune Enabled | <span style=\"color:red\">**" + str(domain_config.auto_tune_enable) + "**</span> |")
    if domain_config.encryption_at_rest:
        write_line(report_file, "| Security | Encryption At Rest | " + str(domain_config.encryption_at_rest) + " |")
    else:
        write_line(report_file, "| Security | Encryption At Rest | <span style=\"color:red\">**" + str(domain_config.encryption_at_rest) + "**</span> |")
    if domain_config.node2node_encryption:
        write_line(report_file, "| Security | Node to Node Encryption | " + str(domain_config.node2node_encryption) + " |")
    else:
        write_line(report_file,
                   "| Security | Node to Node Encryption | <span style=\"color:red\">**" + str(domain_config.node2node_encryption) + "**</span> |")
    if domain_config.enforce_https:
        write_line(report_file, "| Security | HTTPS Enabled | " + str(domain_config.enforce_https) + " |")
    else:
        write_line(report_file, "| Security | HTTPS Enabled | <span style=\"color:red\">**" + str(domain_config.enforce_https) + "**</span> |")
    if domain_config.fgac:
        write_line(report_file, "| Security | Fine Grained Access Control | " + str(domain_config.fgac) + " |")
    else:
        write_line(report_file, "| Security | Fine Grained Access Control | <span style=\"color:red\">**" + str(domain_config.fgac) + "**</span> |")
    write_line(report_file, "| Security | Internal User Database | " + str(domain_config.internal_user_database) + " |")
    #write_line(report_file, "| Access Policy: | " + domain_config.access_policy + " |")
    write_line(report_file, "")

def output_cluster_settings_and_state(data_folder, recommendations,report_file):
    pass


def output_cluster_health_and_stats(data_folder,recommendations, report_file):
    cluster_health_recommendations, cluster_health = gen_cluster_health(data_folder, report_file)
    cluster_stats_recommendations, cluster_stats = gen_cluster_stats(data_folder, report_file)
    write_line(report_file, "### Cluster Status Information:")
    write_line(report_file, "")
    write_line(report_file,
               "| Category |  Setting  |                        Value                        |")
    write_line(report_file,
               "|:------: | :----:| :-------------------------------------------------------: |")
    write_line(report_file, "| Health | Name | " +cluster_health["cluster_name"] + " |")
    write_line(report_file, "| Health | Status | " + cluster_health["status"] + " |")
    if cluster_health["relocating_shards"] > 0:
        write_line(report_file, "| Health | relocating_shards | <span style=\"color:red\">**" + str(cluster_health["relocating_shards"]) + "**</span> |")
    if cluster_health["initializing_shards"] > 0:
        write_line(report_file, "| Health | initializing_shards | <span style=\"color:red\">**" + str(cluster_health["initializing_shards"]) + "**</span> |")
    if cluster_health["unassigned_shards"] > 0:
        write_line(report_file, "| Health | unassigned_shards | <span style=\"color:red\">**" + str(cluster_health["unassigned_shards"]) + "**</span> |")
    if cluster_health["delayed_unassigned_shards"] > 0:
        write_line(report_file, "| Health | delayed_unassigned_shards | <span style=\"color:red\">**" + str(cluster_health["delayed_unassigned_shards"]) + "**</span> |")
    if cluster_health["number_of_pending_tasks"] > 0:
        write_line(report_file, "| Health | number_of_pending_tasks | <span style=\"color:red\">**" + str(cluster_health["number_of_pending_tasks"]) + "**</span> |")
    if cluster_health["task_max_waiting_in_queue_millis"] > 0:
        write_line(report_file, "| Health | task_max_waiting_in_queue_millis | <span style=\"color:red\">**" + str(cluster_health["task_max_waiting_in_queue_millis"]) + "**</span> |")

    write_line(report_file, "| Stats | Total Documents | " + str(cluster_stats["indices"]["docs"]["count"]) + " |")
    write_line(report_file, "| Stats | Store Size | " + str(cluster_stats["indices"]["store"]["size"]) + " |")
    if cluster_stats["indices"]["fielddata"]["evictions"] > 0:
        write_line(report_file, "| Stats | Field Data Memory | " + str(cluster_stats["indices"]["fielddata"]["evictions"]) + " |")
    query_cache_total = cluster_stats["indices"]["query_cache"]["total_count"]
    query_cache_miss = cluster_stats["indices"]["query_cache"]["miss_count"]
    if query_cache_total > 0:
        write_line(report_file,
                   "| Stats | Query Cache Miss | " + "{:.2f}".format(query_cache_miss/query_cache_total * 100) + "% |")
    query_cache_cache_size = cluster_stats["indices"]["query_cache"]["cache_size"]
    if query_cache_cache_size > 0:
        write_line(report_file, "| Stats | Query Cache Size | " + str(query_cache_cache_size) + " |")
    query_cache_evictions = cluster_stats["indices"]["query_cache"]["evictions"]
    if query_cache_evictions > 0:
        write_line(report_file, "| Stats | Query Cache Evictions | " + str(query_cache_evictions) + " |")

    field_types = cluster_stats["indices"]["mappings"]["field_types"]
    for ft in field_types:
        write_line(report_file, "| Mapping Field | " +ft["name"] + " | Total fields/Total indices:" + str(ft["count"]) +
                   "/" + str(ft["index_count"]) + "  |")


def output_node(data_folder, recommendations, report_file):
    write_line(report_file, "## Nodes")
    write_line(report_file, "")
    #nodes_recommendations, nodes_data = gen_cluster_nodes(data_folder, report_file)
    output_data_nodes_table(data_folder, report_file)
    write_line(report_file, "")
    write_line(report_file, "### Recommendations")
    write_line(report_file, "")
    output_body_section(recommendations,report_file,"node")
    write_line(report_file, "")

def output_index(data_folder, recommendations, report_file):
    write_line(report_file, "## Index")
    write_line(report_file, "")
    write_line(report_file, "")
    index_recommendations, index_data = gen_index(data_folder, report_file)
    domain_recommendations, domain_config = gen_domain_config(data_folder, report_file)
    total_index =len(index_data)
    total_primary_size = index_data["pri.store.size"].sum()
    total_size = index_data["store.size"].sum()
    write_line(report_file,
               "| Attribute |                        Value                        |")
    write_line(report_file,
               "|:------: | :------------------------------------: |")
    write_line(report_file, "| Total Index | " + str(total_index) + " |")

    write_line(report_file, "| Primary Volume| " + "{:.1f}".format(total_primary_size/(1024*1024*1024)) + "(GB) |")
    write_line(report_file, "| Replica Volume | " + "{:.1f}".format((total_size-total_primary_size)/(1024*1024*1024)) + "(GB) |")
    write_line(report_file, "")

    index_data = index_data[~(index_data["index"].str.startswith(".kibana",na=False))]
    top_20_index = index_data.sort_values("store.size",ascending=False).head(20)
    top_20_index = top_20_index[["index","status","pri","rep","store.size","pri.store.size","docs.count"]]
    bottom_20_index = index_data.sort_values("store.size", ascending=True).head(20)
    bottom_20_index = bottom_20_index[["index","status","pri","rep","store.size","pri.store.size","docs.count"]]

    write_line(report_file, "### Largest 20 Index")
    write_line(report_file, "")
    gen_table_appendix(report_file, top_20_index)
    write_line(report_file, "")
    write_line(report_file, "### Smallest 20 Index")
    write_line(report_file, "")
    gen_table_appendix(report_file, bottom_20_index)

    write_line(report_file, "### Recommendations")
    write_line(report_file, "")
    output_body_section(recommendations, report_file, "index")
    write_line(report_file, "")

def output_shard(data_folder, recommendations, report_file):
    write_line(report_file, "## Shards")
    write_line(report_file, "")
    shard_recommendations, shard_data = gen_shard(data_folder, report_file)

    shard_data = shard_data[~(shard_data["index"].str.startswith(".", na=False))]
    top_20_shard = shard_data.sort_values("store", ascending=False).head(20)
    top_20_shard = top_20_shard[["index", "shard", "prirep", "state", "store", "docs"]]
    bottom_20_shard = shard_data.sort_values("store", ascending=True).head(20)
    bottom_20_shard = bottom_20_shard[["index", "shard", "prirep", "state", "store", "docs"]]

    primary_size = len(shard_data[shard_data["prirep"] == "p"])
    replica_size = len(shard_data[shard_data["prirep"] == "r"])
    write_line(report_file,
               "| Attribute |                        Value                        |")
    write_line(report_file,
               "|:------: | :------------------------------------: |")
    write_line(report_file,
               "| Number of Primary Shard | " + str(primary_size) + " |")
    write_line(report_file,
               "| Number of Replica Shard | " + str(replica_size) + " |")
    write_line(report_file,
               "| Max Replica size | " + str(top_20_shard["store"].tolist()[0]) + " |")
    write_line(report_file,
               "| Min Replica size | " + str(bottom_20_shard["store"].tolist()[0]) + " |")
    if shard_data["store"].mean() > 20000000000:
        write_line(report_file,
               "| Avg Replica size | <span style=\"color:red\">**{:.1f} | ".format(shard_data["store"].mean()) + "**</span> |")
    else:
        write_line(report_file,
                   "| Avg Replica size | {:.1f} | ".format(shard_data["store"].mean()) + " |")
    write_line(report_file,
               "| Replica size larger than 20 GB | <span style=\"color:red\">**" + str(len(shard_data[shard_data["store"]>20000000000])) + "**</span> |")
    write_line(report_file,
               "| Replica size larger than 20 GB(part of the index) | <span style=\"color:red\">**" + str(shard_data[shard_data["store"] > 20000000000]["index"].unique().tolist()[:20]) + "**</span> |")
    write_line(report_file,
               "| Replica size smaller than 1 GB | <span style=\"color:red\">**" + str(len(shard_data[shard_data["store"] < 1000000000])) + "**</span> |")
    write_line(report_file,
               "| Replica size larger than 1 GB(part of the index) | <span style=\"color:red\">**" + str(
                   shard_data[shard_data["store"] < 1000000000]["index"].unique().tolist()[:20]) + "**</span> |")
    write_line(report_file, "")


    write_line(report_file, "### Largest 20 Shard")
    write_line(report_file, "")
    gen_table_appendix(report_file, top_20_shard)
    write_line(report_file, "")
    write_line(report_file, "### Smallest 20 Shard")
    write_line(report_file, "")
    gen_table_appendix(report_file, bottom_20_shard)
    write_line(report_file, "### Recommendations")
    write_line(report_file, "")
    output_body_section(recommendations, report_file, "shard")
    write_line(report_file, "")


def output_body_section(recommendations, report_file,section):
    #print("output_body_section" + str(recommendations))
    priority_recommendation={"1":[],"2":[],"3":[],"4":[],"5":[]}
    for r in recommendations:
        rec = recommendations[r]["recommendation"]
        if rec.lower() != "No Recommendation".lower():
            report_section = recommendations[r]["report_section"]
            if report_section.lower() == section.lower():
                rule_name = recommendations[r]["rule_name"]
                priority = recommendations[r]["priority"]
                observation = recommendations[r]["observation"]
                impact = recommendations[r]["impact"]
                rec_text = recommendations[r]["recommendation"]
                #field = recommendations[r]["field"]
                #original = recommendations[r]["original"]
                rec_entry=(impact,rec_text)
                priority_recommendation[str(priority)].append(rec_entry)

    write_line(report_file,
               "| Priority |  Category  |                        Recommendation               |")
    write_line(report_file,
               "|:------: | :----:| :-------------------------------------------------------: |")
    for p in priority_recommendation:
        priority_rec_list = priority_recommendation[p]
        for rec_entry in priority_rec_list:
            category = rec_entry[0]
            recommendation = rec_entry[1]
            write_line(report_file, "| {} | {} | {} |".format(p,category,recommendation))


def output_thread_pool_table(data_folder, report_file):
    write_line(report_file, "### ThreadPool")
    with open(data_folder + "/thread_pool_table.csv") as f:
        data = pd.read_csv(f, delim_whitespace=True)
        data = data[(data['queue'] > 0) | (data['rejected'] > 0)]
        data.sort_values(by=['rejected', 'queue'], ascending=False, inplace=True)
        gen_table_appendix(report_file, data)


def output_allocation_table(data_folder, report_file):
    write_line(report_file, "### Allocation")
    with open(data_folder + "/allocation_table.csv") as f:
        data = pd.read_csv(f, delim_whitespace=True)
        data.sort_values(by=['shards', 'disk.indices', 'disk.used'], ascending=False, inplace=True)
        gen_table_appendix(report_file, data)


def output_shards_table(data_folder, report_file):
    write_line(report_file, "### Shard")
    with open(data_folder + "/shards_detail_table.csv") as f:
        data = pd.read_csv(f, delim_whitespace=True)
        data.sort_values(by=['store', 'docs'], ascending=False, inplace=True)
        gen_table_appendix(report_file, data)


def output_index_table(data_folder, report_file):
    write_line(report_file, "### Index")
    with open(data_folder + "/indices_table.csv") as f:
        data = pd.read_csv(f, delim_whitespace=True)
        data.sort_values(by=['store.size', 'docs.count'], ascending=False, inplace=True)
        gen_table_appendix(report_file, data)


def output_data_nodes_table(data_folder, report_file):
    write_line(report_file, "### DataNodes")
    write_line(report_file,"")
    with open(data_folder + "/nodes_cat_table.csv") as f:
        data = pd.read_csv(f, delim_whitespace=True)
        data.sort_values(by=['heap.percent', 'load_15m'], ascending=False, inplace=True)
        gen_table_appendix(report_file, data)


def output_recommendations_as_csv(recommendations,output_folder):
    with open(output_folder+"/recommendations.csv","a+", encoding='utf-8') as f:
        f.write("rule_name,priority,observation,impact,report_section\n")
        for r in recommendations:
            if recommendations[r]["recommendation"].lower() != "No Recommendation".lower():
                rule_name = recommendations[r]["rule_name"]
                #print("rule name:" + rule_name)
                priority = recommendations[r]["priority"]
                observation = recommendations[r]["observation"]
                impact = recommendations[r]["impact"]
                report_section = recommendations[r]["report_section"]
                rec_text = "\"{0}\",{1},\"{2}\",\"{3}\",\"{4}\"".format(rule_name,priority,observation,impact,report_section)
                f.write(rec_text)
                f.write("\n")


def read_recommendations_from_csv(output_folder):
    rec_df = pd.read_csv(output_folder+"/recommendations.csv")
    rec_df = rec_df[(rec_df['priority'] == 1)]
    rec_df.sort_values(by=["report_section"],ascending=True,inplace=True)
    return rec_df


